-- %Version EB Dec 12















-- 
-- In this section we analyse the given Light Control requirements,
-- largely following the order in which they are presented in the given
-- document which we will call for simplicity also Problem Description,
-- and express them in a mathematical form. We introduce formal pendants
-- (the signature) for the various objects and operations mentioned in
-- the document and use definitions and algorithmic notation (ASM rules)
-- to express their static and their dynamic dependencies.
-- 
-- The given document starts with the floor description (Part 2) which
-- describes the signature of the system to be developed, i.e. a listing
-- of its components with the associated data structures.
-- We reflect this by listing and
-- classifying the basic system (classes of) objects and their
-- properties. In Part 3 the Problem Description distinguishes three
-- categories of informal needs, the user needs, the facility manager
-- needs and fault tolerance. We obtain a modularization of the
-- requirements by parameterizing and grouping the user and the facility
-- manager needs into the requirements for the possible manual
-- interactions of the user/facility manager with the control system,
-- such as pressing buttons (reflecting essentially section 3.1. of the
-- Problem Description), and into the automatic actions which are
-- triggered by the control system, such as switching off light in
-- unoccupied rooms. At the end we describe the malfunctions (reflecting
-- section 3.2. of the document).
-- 
-- This is already a good place to introduce the formal representation of
-- a distinction which is made in the Problem Description for each
-- category of needs, namely between actions which are triggered by the
-- user or the facility manager and actions which are triggered by the
-- control system (due to some calculation on the basis of external
-- information). We express this distinction by a function $mode$ which
-- indicates whether the current light setting operation in a room has
-- been done manually, by the user or the facility manager, or
-- automatically. In an object-oriented perspective this function can be
-- thought of as being parameterized by rooms.
-- 
-- \subsection{Basic objects and operations (Signature)} The basic
-- objects appearing in the floor description are rooms and hallways
-- (more precisely hallway sections). They are associated with light
-- groups (window and wall ceiling lights for rooms and ceiling lights
-- for hallways) which come with operations of pushing various buttons
-- (on the wall or a control panel) and of actuating dimmers. Rooms and
-- hallways are also associated with various motion detectors and light
-- sensors and  door closing contacts. There are also
-- status lines which report status values of the associated light
-- groups. 
-- 
-- The staircases which are mentioned in the Floor Description
-- enter the problem really only through their motion detector. To avoid
-- the proliferation of irrelevant object types we include these
-- staircase motion detectors in the class of motion detectors which are
-- related to the doors for entering a hallway from a staircase.
-- 
-- These objects enter our formalization as parameters of the various
-- actions which are described in the following sections. In a systematic
-- documentation of the requirements elicitation one has to list
-- explicitly the complete signature; the basic objects have to be
-- defined through the lexicon, their properties as well as the
-- conditions imposed on the operations have to be listed systematically,
-- making sure that the list is complete and correct with respect to the
-- application domain information which underlies the Problem
-- Description. We abstain from doing this here; any reliable,
-- systematic and computer assisted method can serve this purpose.
-- 
-- 
-- \subsection{User interactions (manual light control)}
--    \label{Sec:ManualControl}
-- 
-- The Problem Description states in section 2.1-2.4 (paragraphs 7-11)
-- that every room has two wall switches, one for the wall ceiling
-- light group and the other for the window ceiling light group. The
-- behavior is formulated in 2.1.2 as follows.
-- 
--  \begin{enumerate}
--   \item If the ceiling light group is completely on, it will be
--         switched off,
--   \item otherwise it will be switched on completely.  
-- \end{enumerate}
-- 
-- We capture this requirement with the following rule
-- $room\_wall\_button$. This rule is parameterized over a room and a
-- light group.  The fact that pressing the button is an external event
-- over which the system has no control is reflected by an event function
-- $lightgroup\_wall\_button\_pressed$ which is supposed to become
-- $\constructor{True}$ when the corresponding button has been pressed
-- and to become $\constructor{False}$ by firing the rules where the
-- event function appears in the guard (\PushButtonReq). This
-- interpretation resolves the incompleteness of the definition for ``push
-- button'' in the dictionary where it is not clear when the
-- light effect shoud take place, at the begin or at the end of the
-- possibly prolonged button pushing action.
-- 
-- Here and in the following we name
-- such additional requirements, for the purpose of a systematic
-- documentation of all the decisions taken to interpret or extend the
-- requirements in the Problem Description.
-- 
-- \begin{asm}
           
room_wall_button :: (Room, LightGroup) -> Rule ()
room_wall_button (room, lightgroup) =
    if lightgroup_wall_button_pressed(room, lightgroup) then
      if lightgroup_is_completely_on(room, lightgroup) then
        switch_lightgroup_off(room, lightgroup)
       else
        switch_lightgroup_completely_on(room, lightgroup)
    else skip 
-- \end{asm}
-- 
-- \noindent
-- Switching a light group off and switching it completely on is defined
-- as setting all lights in the corresponding room to
-- $\constructor{minDimValue}$ or $\constructor{maxDimValue}$
-- respectively. We use these symbolic names rather than constants
-- like $0\%$ and $100\%$.
-- 
-- 
-- 
-- As explained above the dynamic function $mode$ determines for each
-- room whether the light was set by the user ($\constructor{Manual}$) or
-- by the control system ($\constructor{Ambient}$).
-- 
-- \begin{asm}
           
switch_lightgroup_off :: (Room, LightGroup) -> Rule ()
switch_lightgroup_off (room, lightgroup) = do
   mode(room) := Manual
   forall light <- lights_in_group(room, lightgroup) do
      switch_light(room, light, minDimValue)
-- \end{asm}
-- 
-- 
-- \begin{asm}
           
switch_lightgroup_completely_on :: (Room, LightGroup) -> Rule ()
switch_lightgroup_completely_on (room, lightgroup) = do
   mode(room) := Manual
   forall light <- lights_in_group(room, lightgroup) do
      switch_light(room, light, maxDimValue)
-- \end{asm}
-- 
-- \noindent Following 2.6  in every
-- hallway section there are switch buttons, linked in parallel (see
-- 2.6.3). The light in the hallway section has to be on if one button is
-- defect ($any\_hallway\_button\_defect$). This is according to {\tt
-- NF5} where we interprete ``not controllable manually'', in view of the
-- safety requirement {\tt U1}, as meaning that at least one (not
-- necessarily every) hallway button is defect (\NFfiveAReq). It would
-- be easy in our ASM model to change this to a ``local'' interpretation
-- of ``not controllable manually'' in {\tt NF5} by parameterizing
-- $any\_hallway\_button\_defect$ by buttons as arguments. The event
-- function $hallway\_button\_pressed$ indicates that a switch button has
-- been pressed and becomes false by firing the rules where it appears as
-- guard.
-- 
-- \begin{asm}
           
hallway_button :: Hallway -> Rule ()
hallway_button (hallway) =
    if hallway_button_pressed(hallway) &&
       not (any_hallway_button_defect(hallway)) then 
     if light_is_on(hallway) then
       switch_lights_off(hallway)
     else 
       switch_lights_on(hallway)
    else skip 
-- \end{asm}
-- 
-- \noindent
-- According to the dictionary definition of ``push button'', switching
-- light manually on or off in rooms and hallways is similar. We reflect
-- this uniformity by introducing the term {\em location} standing for rooms
-- and hallways.  Switching on/off for a location is defined as setting
-- all lights of the location to minDimValue or maxDimValue
-- respectively. We group these lights for short as
-- $lights\_at(location)$. Although for uniformity reasons we formulate
-- $switch\_lights\_on$ for locations, we will use $switch\_lights\_on$ only
-- for hallways, because by requirement 8.2, {\tt U5,U6,U9} and the
-- dictionary entry ``light scene'', the light in a room is only
-- switched on for a lightgroup.
-- 
-- \begin{asm}
           
switch_lights_on :: Location -> Rule ()
switch_lights_on (location) =
   forall light <- lights_at(location) do
      switch_light(location, light, maxDimValue)
-- \end{asm}
-- 
-- \begin{asm}
           
switch_lights_off :: Location -> Rule ()
switch_lights_off (location) = do
   forall light <- lights_at(location) do
      switch_light(location, light, minDimValue)
   if location_is_room(location) then do
     mode(location) := Manual
    else skip 
-- \end{asm}
-- 
-- 
-- \noindent The facility manager can switch off the ceiling light
-- in rooms and in hallway sections if the corresponding room or hallway
-- is not occupied ({\tt FM6}).
-- 
-- \begin{asm}
           
manually_switch_off :: Location -> Rule ()
manually_switch_off (location) =
    if manually_switch_off_pressed(location) && not (occupied(location)) then
      switch_lights_off(location)
    else skip 
-- \end{asm}
-- 
-- \noindent The rule uses the function $occupied$. The Problem
-- Description doesn't describe how to determine occupation. Rooms and
-- hallways have motion sensors but these can only sensor motion. Imagine
-- somebody for a while sits quietly on his chair in a room so that the
-- motion sensor reports absence of motion. But the room is still
-- occupied. Therefore a reasonable definition for a location to be not
-- occupied is that there has been no motion for a period of
-- $max\_quiet\_time$ (\RoomOccupationReq). It remains to be determined
-- whether this function is fixed once and for all or whether it can be
-- changed and what is its concrete value for a given system. To reflect
-- the malfunction requirement {\tt NF4} we include the case that a
-- location is occupied if at least one of its motion detectors does not
-- work correctly, so that in this case the light cannot be switched off
-- by the facility manager. To guarantee the consistency between user and
-- facility manager light updates by the rules $room\_wall\_button$,
-- $hallway\_button$, $manually\_switch\_off$,
-- $control\_panel$ we assume that the motion sensor detects
-- when users push buttons (\MotionDetectorReq). This is a constraint
-- which relates the notion of being occupied to the event functions
-- $pressed$ associated to buttons.
-- 
-- \begin{asm}
           
occupied :: Location -> Bool
occupied (location) = 
   current_time - last_motion(location) <= max_quiet_time
-- \end{asm}
-- 
-- \noindent
-- The dynamic function $last\_motion$ stores the time of the last motion
-- and we update the function by observing the motion detector.  The
-- monitored function $somebody\_is\_moving$ yields the value of the
-- corresponding motion detector. To satisfy {\tt NF5} we assume that the
-- function $somebody\_is\_moving$ is $\constructor{True}$ if the
-- corresponding motion detector is defect (\NFfiveBReq).
-- 
-- \begin{asm}
           
observe_motion_detector :: Location -> Rule ()
observe_motion_detector (location) =
    if somebody_is_moving(location) then
       last_motion(location) := current_time
    else skip 
-- \end{asm}
-- 
-- \noindent There is one more action which is triggered by user
-- interaction. Following {\tt U5,U6,U9} one can use the control panel
-- to control the ceiling lights and the light scene.  The ceiling lights
-- can be switched on and off. Before collecting the requirements for the
-- light scene we first analyse the overall functionality of the control
-- panel. We use an additional event function $switch\_value$ to express
-- the on/off position chosen by the user for the switch in question.
-- 
-- \begin{asm}
           
control_panel :: (Room, PanelSwitch) -> Rule ()
control_panel (room, switch) = 
  if switch_pressed(room, switch) then case switch of
       AmbientSelection -> 
         activate_light_scene(room, last_light_scene(room))
       LightGroup(lg) -> 
         case switch_value(room, switch) of
            On  -> switch_lightgroup_completely_on(room, lg)
            Off -> switch_lightgroup_off(room, lg)
       SceneSelection   -> 
         case switch_value(room, switch) of
          Scene(s) -> set_light_scene(room,s)
     else skip 
-- \end{asm}
-- 
-- \noindent
-- One has to guarantee the consistency of pushing on wall buttons or on
-- the control panel. One can for example assume that the hardware solves
-- this conflict, or one could establish a fixed priority (\PushButtonReq).
-- 
-- The preceding definition fulfills the informal needs {\tt
-- U5}, {\tt U6} and {\tt U9}.  The button
-- $\constructor{AmbientSelection}$ activates the light scene which was
-- set with $\constructor{SceneSelection}$.
-- 
-- \begin{asm}
           
set_light_scene :: (Room, Scene) -> Rule ()
set_light_scene(room, scene) =
   if mode(room) == Ambient then
     activate_light_scene(room, scene)
    else
     last_light_scene(room) := scene
-- \end{asm}
-- 
-- \noindent 
-- The rule $set\_light\_scene$ for scene selection in
-- ambient mode changes the current activated light scene and in manual
-- mode simply stores the selected value in the dynamic function
-- $last\_light\_scene$. In the second case the scene can be activated by
-- pressing $\constructor{AmbientSelection}$.
-- 
-- A light scene contains an {\em ambient light level} and an ordered
-- list of lights together with a dim value for each light (see
-- requirement 2.10, paragraph 19). As the dictionary indicates under the
-- entry light scene, the control system has to switch on the lights in
-- the given order with the corresponding dim value in order to achieve
-- the specified ambient light level. Reflecting {\tt FM1} the control
-- system must also take into account the ambient light from outside.  We
-- capture these requirements by introducing a function
-- $lights\_to\_turn\_on$ which computes an ordered set containing all
-- lights that should be switched on in this order together with their
-- dim values (\LightSceneReq). Introducing an order makes the dictionary
-- definition of ``light scene'' uniform with respect to the way light
-- scenes and their light groups are built from components, achieving
-- easy adaptability to changing requirements. The function depends for
-- each room on the value of the outdoor light sensor and of the
-- activated light scene. The fact that we use this function only for
-- rooms and not for hallways reflects that the requirements {\tt
-- FM1} and {\tt NF3} are useless for hallways if they have no windows, as is
-- suggested by Fig.1 in paragraph 5 of the Problem Description 
-- (\HallwayReq). Note that the monitored function $lights\_to\_turn\_on$ 
-- (see below) takes 
-- into account the information about malfunctioning lights.  
-- 
-- 
-- \begin{asm}
           
activate_light_scene :: (Room, Scene) -> Rule ()
activate_light_scene (room, scene) = do
    mode(room)             := Ambient
    last_light_scene(room) := scene

    if scene == default_light_scene(room) &&
       outdoor_light_sensor_defect(room) then
         switch_lights_on(room)
     else do
        let lights_on = lights_to_turn_on(room,outdoor_sensor(room), scene) 
        forall (light, value) <- lights_on do
          switch_light( room , light, value )
        forall light <- lights_at(room) `listDifference` [  l | (l,v) <- lights_on  ]  do
         switch_light(room, light, minDimValue)
-- \end{asm}
-- 
-- \noindent
-- This rule also correctly reflects the requirement {\tt NF2}: 
-- 
-- \begin{quote} 
--   If any outdoor light sensor does
--   not work correctly, the default light scene for alls rooms is that
--   both ceiling light groups are on.
-- \end{quote}
-- 
-- \noindent
-- Also the part of {\tt NF1} which complements {\tt NF2} is reflected by
-- the assumption that the value of $outdoor\_sensor(room)$ remains
-- constant if the sensor does not work correctly
-- (\OutdoorSensorReq). This assumption reflects that {\tt NF1} is not a
-- requirement on the controller, but on the way the sensor values are
-- transmitted as input to the controller.
-- 
-- \noindent
-- Anticipating the rule $use\_daylight$ below we point out already here
-- that the definition of $activate\_light\_scene$ contains a decision about
-- the interpretation of requirement {\tt U10}. Since nothing is said
-- about what it means to maintain the ceiling light group on a
-- given light scene, 
-- we interpret this as requesting that the ceiling lights
-- are set to $\constructor{minDimValue}$ if they do not enter explicitly
-- the lights to
-- be turned on for the given light scene (\UtenReq).
-- 
-- The control panel also allows to set the value T1 for a room 
-- ({\tt U7}).  This is formalizable by updates of a corresponding function
-- $t_1$, similarly for {\tt FM4,FM5}, which we group into the 
-- $set\_parameters$ part of the $failure\_and\_service$ submachine.
-- 
-- 
-- \subsection{Automatic light control}
--    \label{Sec:AutomaticControl}
-- 
-- The automatic control system is required to be able to switch on and
-- off any light in a room or in a hallway section. Switching on is used
-- to ensure that there is safe illumination in the corresponding room or
-- hallway ({\tt U1}, {\tt U13}, {\tt U14}). The lights in the hallway
-- sections are not dimmable so that switching on can be done there only
-- completely. Switching on is triggered by the two events (i) motion in
-- the hallway ($somebody\_is\_moving$) and (ii) a door is open
-- ($some\_door\_is\_open$).
-- 
-- \begin{asm}
           
auto_switch_on_in_hallway :: Hallway -> Rule ()
auto_switch_on_in_hallway (hallway) =
    if (somebody_is_moving(hallway) || some_door_is_open(hallway)) &&
       light_is_off(hallway)
    then 
      switch_lights_on(hallway)
    else skip 
-- \end{asm}
-- 
-- \noindent
-- Switching on the light in a room is more complicated. Following {\tt U3}
-- and {\tt U4} one has to
-- distinguish two cases:
-- \begin{description}
--   \item[{\tt U3}]
--         If the room is reoccupied  within T1 minutes after the
--         last person has left the room, the {\em chosen light scene}
--         has to be reestablished.
--   \item[{\tt U4}]
--         If the room is reoccupied after more than T1 minutes since
--         the last person has left the room, the {\em default
--         light scene} has to be established.
-- \end{description}
-- 
-- \noindent In the first case, instead of establishing the {\em chosen
-- light scene} we use the {\em last light scene} (\UthreeReq) since otherwise
-- the requirements would be incoherent, as the following example
-- shows. By the definition in the dictionary in Part 4 of the Problem
-- Description, the {\em chosen light scene} is the scene selected with the
-- control panel. Imagine the following scenario: 
-- 
-- \begin{enumerate}
--   \item Person $A$ enters the room and selects scene $s$.
--   \item The person leaves the room for more than T1 minutes.
--   \item Person $B$ enters the room. According to  {\tt U4}
--         the {\em default light scene} should be established.
--   \item Person $B$ does not change the light scene and leaves the room.
--   \item Person $B$ enters within T1 minutes. According to {\tt U3}
--         we should establish the {\em chosen light 
--         scene} $s$.
-- \end{enumerate}
-- 
-- \noindent
-- As one can see Person $B$, upon entering the room for the
-- first time, gets the {\em default light scene} and, upon reentering,
-- gets the {\em chosen light scene} of Person $A$. This seems to be a
-- flaw in the Problem Description and we therefore select the {\em last
-- light scene}.  For the reasons explained above we abstract from the
-- value T1 and use the function $recently\_occupied$.
-- 
-- \begin{asm}
           
auto_switch_on_in_room :: Room -> Rule ()
auto_switch_on_in_room (room) =
    if (somebody_is_moving(room) || some_door_is_open(room)) &&
       light_is_off(room)
    then 
      activate_light_scene(room, scene)
    else skip 
    where scene = if recently_occupied(room) then 
                    last_light_scene(room)
                  else default_light_scene(room) 

recently_occupied :: Room -> Bool
recently_occupied (room) = current_time - last_motion(room) <= t1(room)
-- \end{asm}
-- 
-- 
-- \noindent We do not commit here to any particular definition of
-- $default\_light\_scene$ (\DefaultLightSceneReq). The definition in the
-- dictionary is probably not reasonable because with that definition
-- requirement {\tt U4} makes no sense.  
-- 
-- The control system switches off
-- the light in a room or in a hallway section if the location is not
-- occupied for T3 or T2 minutes respectively ({\tt FM2}, {\tt
-- FM3}). In accordance with requirement {\tt NF5} we don't switch off
-- the light in a hallway section if one of its buttons is defect. To
-- reflect the malfunction condition {\tt NF4} we stipulate that
-- occurrence of a malfunction for a motion sensor is interpreted as
-- presence of motion so that the location appears as occupied.
-- 
-- 
-- 
-- 
-- \begin{asm}
           
auto_switch_off :: Location -> Rule ()
auto_switch_off (location) =
    if location_is_hallway(location) && any_hallway_button_defect(location) 
     then skip
     else 
      if not(occupied(location)) && no_motion_for_long_time(location) &&
         not(some_door_is_open(location)) && not(light_is_off(location))
      then switch_lights_off(location)
      else skip 

no_motion_for_long_time :: Location -> Bool
no_motion_for_long_time (location) = 
      if location_is_room(location) then
        current_time - last_motion(location) > t3(location)
      else
        current_time - last_motion(location) > t2(location)
-- \end{asm}
-- 
-- \noindent
-- The control system should use daylight to achieve the desired {\em ambient
-- light level} ({\tt FM1}). We model this by reactivating the current
-- light scene if the room is in ambient mode and there
-- is no request for the ceiling lights. The following rule also
-- reflects the informal need {\tt U10}.
-- 
-- \begin{asm}
           
use_daylight :: Room -> Rule ()
use_daylight (room) =
    if no_event_for_ceiling_light(room) && mode(room) == Ambient then
     activate_light_scene(room, last_light_scene(room))
    else skip 
-- \end{asm}
-- 
-- \noindent
-- {\tt U2} is fulfilled automatically in our requirements model because
-- an ASM state remains unchanged unless a specific (user or control
-- system) action triggers a change for certain state components
-- (arguments where functions may change their value). 
-- 
-- \subsection{Failure and service}
--    \label{Sec:Failure}
-- 
-- The last part of the Problem Description is about malfunctions. There
-- are two problems which occur here, namely identifying and handling
-- malfunctions. Identifying malfunctions is a rather difficult
-- application domain and not so much a software design problem. The
-- Problem Description does not provide any further details on this issue
-- so that we assume having a function $malfunction\_occurs$ telling
-- whether a component works correctly or not; a component may be a hallway
-- button, a light sensor, a motion sensor or any light.
--  For building a concrete plant with its control software, this
-- function has to be further specified by the customer, together
-- with the support requested in {\tt FM8} for finding the reasons
-- for occuring malfunctions. To reflect the
-- malfunction requirement {\tt FM1} we stipulate that this function can
-- also be updated manually.
-- 
-- 
-- \begin{asm}
           
malfunction :: Rule ()
malfunction =
   forall component <- all_components do
     if malfunction_occurs(component) then
       handle_malfunction(component)
      else skip 
-- \end{asm}
-- 
-- \noindent According to {\tt U8}, {\tt FM7} and {\tt FM10} the handling
-- of malfunction logs the corresponding information. 
-- In the case where a hallway button is defect we
-- switch on the lights in that hallway ({\tt NF5}). In case a
-- hallway motion detector is defect, by assumption (\NFfiveBReq) the
-- function $somebody\_is\_moving$ is true and we therefore switch on
-- the lights by rule $auto\_switch\_on\_in\_hallway$.
-- 
-- \begin{asm}
           
handle_malfunction :: Component -> Rule ()
handle_malfunction (component) =
  case component of
    OutdoorLightSensor(i) ->
        forall room <- rooms_under_lightsensor(i) do
          inform_user(room, LightSensorDefect(i))
          inform_facility_manager(LightSensorDefect(i))
          write_log_in_database(LightSensorDefect(i))
    MotionSensor(location, i) -> do
        if location_is_room(location) then
         inform_user(location, MotionSensorDefect(location,i))
         else skip 
        inform_facility_manager(MotionSensorDefect(location,i))
        write_log_in_database(MotionSensorDefect(location,i))
    HallwayButton(hallway, i) -> do
        switch_lights_on(hallway)
        inform_facility_manager(HallwayButtonDefect(hallway,i))
        write_log_in_database(HallwayButtonDefect(hallway,i))
    Luminaire(location, light) -> do
       inform_facility_manager(LightDefect(location, light))
       write_log_in_database(LightDefect(location, light))
-- \end{asm}
-- 
-- \noindent
-- To reflect requirement  {\tt NF9} we add a rule
-- 

-- 
-- \noindent
-- Since the informal description contains
-- no information on the meaning of {\tt NF9}, we leave it to further
-- refinement steps to provide a detailed definition, resulting from the
-- discussion with the application domain expert who is supposed to know
-- what inputs have to be considered as ``unreasonable''.
-- 
-- The system should provide reports on energy consumption ({\tt FM9}).
-- We formalize this requirement by introducing a dynamic function
-- $dim\_value$ storing the current dim value of a light. Switching the 
-- light is defined as setting a dim value.
-- If the dim value is less than $10\%$ then the light is switched off (see
-- Problem Description, Table 2):
-- \begin{asm}
           
switch_light :: (Location, Light, DimValue) -> Rule ()
switch_light (location, light, value) = do
   if value < 10 then
     status_of_light(location, light)    := Light_Off
    else
     status_of_light(location, light)    := Light_On
   dim_value(location, light)            := value
-- \end{asm}
-- 
-- \noindent
-- Based on the values in $dim\_value$ we can 
-- define the function $power\_consumption$ computing the current power 
-- consumption. The function has to take into account the 
-- malfunctioning of lights:
-- 
-- \begin{asm}
           
power_consumption :: Int

power_consumption = sum  [ p(l,dimValue) | (l, dimValue) <- assocs(dim_value) ]

 where p(l,v) = case light_defect(l) of
                        Nothing      -> c * v
                        Just(Light_On)       -> c * maxDimValue
                        Just(Light_Off)      -> c * minDimValue 
       c = 1
-- \end{asm}
-- We use the constant $c$ to adjust the dim value to the electrical
-- power. The energy consumption is the integral of the power consumption
-- over the time. Therefore we store the power consumption in each
-- step in a dynamic
-- function and define the energy consumption as the product of the
-- interval $t_e$ with the sum of the power consumptions. We assume
-- that the following rule will be executed every $t_e$ minutes.
-- 

-- 
-- 
-- \subsection{The requirements ground model} The model which results
-- from the formalization in the preceding subsections is the following
-- ASM, consisting of three submachines:
-- 
-- \begin{asm}
           
light = do manual_light_control
           automatic_light_control
           failure_and_service 
-- \end{asm}
-- 
-- \noindent
-- The $manual\_light\_control$ actions were described in section
-- \ref{Sec:ManualControl}. We introduced rules for switching light 
-- on/off in hallways
-- and in rooms and described the functionality of the control
-- panel.
-- 
-- \begin{asm}
           
manual_light_control = 
  forall location <- all_locations do
    manually_switch_off(location)
    if location_is_room(location) then do
       room_wall_button(location,LightGroupWall)
       room_wall_button(location,LightGroupWindow)
       control_panel(location,switch(location))
     else skip 
    if not(location_is_room(location)) then
      hallway_button(location)
     else skip 
-- \end{asm}
-- 
-- \noindent
-- The second machine $automatic\_light\_control$ consists of
-- automatically switching light on/off and using daylight to
-- achieve the desired light level. Due to automatically 
-- switching on/off we have also to observe the motion detectors. The
-- rules are described in section \ref{Sec:AutomaticControl}.
-- 
-- \begin{asm}
           
automatic_light_control = do
  forall location <- all_locations do
    auto_switch_off(location)
    observe_motion_detector(location)
    if location_is_room(location) then do
      auto_switch_on_in_room(location)
      use_daylight(location)
     else skip 
    if not(location_is_room(location)) then
      auto_switch_on_in_hallway(location)
     else skip 
-- \end{asm}
-- 
-- \noindent
-- The last of the three submachines is $failure\_and\_service$
-- containing handling of malfunctions, detecting unreasonable inputs,
-- reporting energy consumption, and setting parameters (reflecting
-- {\tt FM11}, {\tt FM4}, {\tt FM5}, {\tt U7}).
-- 
-- \begin{asm}
           
failure_and_service = do
   malfunction
   detect_unresonable_input
   report_energy_consumption
   set_parameters
-- \end{asm}
-- 
-- \noindent
-- One can use different policies for the synchronization of the
-- three machines of $light$
-- which has to guarantee the consistency of their update
-- actions in the shared data area. One possibility is to make certain
-- priority or scheduling assumptions on possibly conflicting actions, as
-- we have indicated at various places during the formalization of the
-- requirements. Another possibility is to impose a concrete scheduling
-- on the coordination of the three submachines. Such a global policy
-- relegates the consistency problem to the local level of the single
-- submachines. For our ground model we can assume for example, as we do
-- for its executable version explained in the next section, that the
-- manual and the automatic submachines alternate at a fixed rate---fast
-- enough to guarantee the desired reaction time of the light system to
-- user or environment input---and that the failure and service
-- submachine is executed in between with a certain predescribed
-- frequency, again determined by the time requirements for failure
-- handling and general services. The Problem Description leaves all
-- these issues completely open so that for the executable version of
-- our ground model we had to take some concrete realistic decisions.
-- 
-- We assume starting in an initial state where
-- all rooms and all hallways are empty and all lights
-- are off. Especially, we assume the following initial values for our
-- dynamic functions:

-- 
-- \section{Ground model validation} 
-- 
-- In this section we provide further details for the macros used in the
-- previous sections which allow us to turn the ground model into an
-- executable model which has been implemented in ASMGofer and is
-- available electronically (see \cite{Schmid99a}). This executable
-- ground model version allows the customer to validate the Problem
-- Description by experiments with our model. A preliminary version of
-- this simulation model has been presented in a demo to the Dagstuhl
-- seminar on Requirements Capture \cite{BoRiSc99}.  The definitions
-- presented in this section add to the requirements which occur in the
-- Problem Description and can be viewed as further decisions made for
-- directing the real (not any more prototypical) design.
-- We skip the standard data structures needed to encode 
-- locations, lights, light groups, actuators, light sensor, etc.
-- 
-- \begin{asm}
           
inform_user :: (Room, Malfunction) -> Rule ()
inform_user (room, malfunction) =
   user_information(room, current_time, malfunction) := True
-- \end{asm}
-- 
-- \begin{asm}
           
inform_facility_manager :: Malfunction -> Rule ()
inform_facility_manager (malfunction) =
   facility_manager(malfunction, current_time) := True
-- \end{asm}
-- 
-- 
-- \begin{asm}
           
write_log_in_database :: Malfunction -> Rule ()
write_log_in_database (malfunction) =
   database(current_time, malfunction) := True
-- \end{asm}
-- 
-- \noindent
-- Typically the physical realization of $inform\_user$
-- will be required to appear on the control panel display, but other
-- solutions are possible.
-- 
-- There are several derived functions which can easily be defined and are
-- needed for an executable version. Some of these definitions
-- are listed below:
-- 
-- \begin{asm}
           
lightgroup_is_completely_on :: (Room, LightGroup) -> Bool
lightgroup_is_completely_on (room, lg) =
    all (\ l  ->   dim_value(room, l) == maxDimValue ) ( lights_in_group(room, lg) ) 
-- \end{asm}
-- 
-- 
-- \begin{asm}
           
light_is_on :: Hallway -> Bool
light_is_on (hallway) =
   all (\ l  ->  status_of_light(hallway,l) == Light_On ) ( lights_at(hallway) ) 
-- \end{asm}
-- 
-- 
-- \begin{asm}
           
light_is_off :: Location -> Bool
light_is_off (location) =
   all (\ l  ->   status_of_light(location,l) == Light_Off ) ( lights_at(location) ) 
-- \end{asm}
-- 
-- 
-- 
-- 
